from typing import Any, Dict, List, Mapping

from sanic.headers import HeaderIterable
from sanic.request import File
from sanic.request import Request as _Request
from sanic.request import RequestParameters

from pait.app.base.adapter.request import BaseRequest, BaseRequestExtend
from pait.util import LazyProperty


class RequestExtend(BaseRequestExtend[_Request]):
    @property
    def scheme(self) -> str:
        return self.request.scheme

    @property
    def path(self) -> str:
        return self.request.path

    @property
    def hostname(self) -> str:
        return self.request.host


class Request(BaseRequest[_Request, RequestExtend]):
    RequestType = _Request
    FormType = RequestParameters
    FileType = File
    HeaderType = HeaderIterable

    def request_extend(self) -> RequestExtend:
        return RequestExtend(self.request)

    def body(self) -> dict:
        return self.request.json

    def cookie(self) -> dict:
        return self.request.cookies

    def header(self) -> HeaderIterable:
        return self.request.headers

    def path(self) -> Mapping[str, Any]:
        return self.request_kwargs

    ##################################################
    # sanic return result like: {"a": [1], "b": [2]} #
    # not support raw_return future                  #
    ##################################################
    @LazyProperty()
    def file(self) -> Dict[str, File]:
        return {key: value[0] for key, value in self.request.files.items()}

    @LazyProperty()
    def form(self) -> dict:
        return {key: value[0] for key, value in self.request.form.items()}

    @LazyProperty()
    def query(self) -> dict:
        return {key: value[0] for key, value in self.request.args.items()}

    @LazyProperty()
    def multiform(self) -> Dict[str, List[Any]]:
        return {key: self.request.form.getlist(key) for key, _ in self.request.form.items()}

    @LazyProperty()
    def multiquery(self) -> Dict[str, Any]:
        return {key: self.request.args.getlist(key) for key, _ in self.request.args.items()}



from datetime import date, datetime
from enum import Enum
from json import JSONEncoder
from typing import Any

from _decimal import Decimal
from pydantic.fields import UndefinedType

from pait.model.template import TemplateVar


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return int(obj.timestamp())
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, TemplateVar):
            obj = obj.get_value_from_template_context()
            if isinstance(obj, UndefinedType):
                obj = None
            return obj
        else:
            return super().default(obj)  # pragma: no cover



from typing import Any

import aiofiles  # type: ignore
from sanic import response as sanic_response
from sanic_testing.testing import SanicTestClient, TestingResponse  # type: ignore

from pait.app.sanic.adapter.response import gen_response, set_info_to_response
from pait.plugin.mock_response import MockPluginProtocol


class MockPlugin(MockPluginProtocol[sanic_response.BaseHTTPResponse]):
    def get_response(self) -> sanic_response.BaseHTTPResponse:
        return gen_response(
            self.pait_response_model.get_example_value(example_column_name=self.example_column_name),
            self.pait_response_model,
        )

    async def async_get_file_response(self, temporary_file: Any, f: Any) -> sanic_response.BaseHTTPResponse:
        await f.write(self.pait_response_model.get_example_value(example_column_name=self.example_column_name))
        await f.seek(0)
        resp = await sanic_response.file_stream(f.name, mime_type=self.pait_response_model.media_type)

        raw_streaming_fn = resp.streaming_fn

        async def _streaming_fn(_response: sanic_response.BaseHTTPResponse) -> None:
            await raw_streaming_fn(_response)
            await temporary_file.__aexit__(None, None, None)

        resp.streaming_fn = _streaming_fn
        return resp

    def set_info_to_response(self, resp: sanic_response.BaseHTTPResponse) -> None:
        set_info_to_response(resp, self.pait_response_model)



from typing import Any, Type

from tornado.web import RequestHandler

from pait.model import BaseResponseModel


def gen_response(response_value: Any, response_model_class: Type[BaseResponseModel], *args: Any, **kwargs: Any) -> Any:
    if response_value is None:
        return None
    tornado_handle: RequestHandler = args[0]
    tornado_handle.set_status(response_model_class.status_code[0])

    if response_model_class.header is not None:
        for k, v in response_model_class.get_header_example_dict().items():
            tornado_handle.set_header(k, v)
    tornado_handle.set_header("Content-Type", response_model_class.media_type)
    tornado_handle.write(response_value)
    return None


def set_info_to_response(tornado_handle: RequestHandler, response_model_class: Type[BaseResponseModel]) -> None:
    tornado_handle.set_status(response_model_class.status_code[0])
    tornado_handle.set_header("Content-Type", response_model_class.media_type)
    if response_model_class.header is not None:
        for k, v in response_model_class.get_header_example_dict().items():
            tornado_handle.set_header(k, v)



from typing import TYPE_CHECKING, Any, Callable, Dict, Type

from pait.model.response import BaseResponseModel, JsonResponseModel
from pait.plugin.base import PluginContext, PrePluginProtocol
from pait.util import get_pait_response_model

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel


class CheckJsonRespPlugin(PrePluginProtocol):
    """Check if the json response result is legal"""

    check_resp_fn: Callable[[Any, PluginContext], None]

    @staticmethod
    def get_json(response_data: Any, context: PluginContext) -> dict:
        raise NotImplementedError()

    @classmethod
    def pre_check_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> None:
        super().pre_check_hook(pait_core_model, kwargs)
        if "check_resp_fn" in kwargs:
            raise RuntimeError("Please use response_model_list param")

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        pait_response_model: Type[BaseResponseModel] = get_pait_response_model(pait_core_model.response_model_list)
        if not issubclass(pait_response_model, JsonResponseModel):
            raise ValueError(f"pait_response_model must {JsonResponseModel} not {pait_response_model}")

        def check_resp_by_dict(response_data: Any, context: PluginContext) -> None:
            if not isinstance(response_data, dict):
                response_data = cls.get_json(response_data, context)
            pait_response_model.response_data(**response_data)  # type: ignore

        kwargs["check_resp_fn"] = check_resp_by_dict
        return kwargs

    def _sync_call(self, context: PluginContext) -> Any:
        response: Any = super().__call__(context)
        self.check_resp_fn(response, context)
        return response

    async def _async_call(self, context: PluginContext) -> Any:
        response: Any = await super().__call__(context)
        self.check_resp_fn(response, context)
        return response

    def __call__(self, context: PluginContext) -> Any:
        if self._is_async_func:
            return self._async_call(context)
        else:
            return self._sync_call(context)



from typing import Any

from pait.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin as _AutoCompleteJsonRespPlugin
from pait.plugin.base import PluginContext

from .unified_response import UnifiedResponsePluginProtocol

__all__ = ["AutoCompleteJsonRespPlugin"]


class AutoCompleteJsonRespPlugin(UnifiedResponsePluginProtocol, _AutoCompleteJsonRespPlugin):
    async def __call__(self, context: PluginContext) -> Any:
        response: Any = await super(AutoCompleteJsonRespPlugin, self).__call__(context)
        return self._gen_response(response, context)



import inspect
from typing import Any, Callable, Dict, Optional

from any_api.openapi import SecurityModelType


class BaseSecurity:
    model: SecurityModelType
    security_name: str

    def pait_handler(self, *args: Any, **kwargs: Any) -> Any:
        pass

    def set_pait_handler(self, func: Callable) -> None:
        if hasattr(func, "pait_handler"):
            raise ValueError("'func' already has pait_handler")  # pragma: no cover
        class_func: Optional[Callable] = getattr(self, "__call__", None)
        if not class_func:
            raise ValueError("class has no __call__")  # pragma: no cover

        func_sig = inspect.signature(func)
        class_func_sig = inspect.signature(class_func)
        for k, v in func_sig.parameters.items():
            if v.annotation != class_func_sig.parameters[k].annotation:
                raise ValueError(f"func parameter {k} annotation not match")  # pragma: no cover
        if func_sig.return_annotation != class_func_sig.return_annotation:
            raise ValueError("func return annotation not match")

        setattr(self, "pait_handler", func)

    @classmethod
    def get_exception(cls, *, status_code: int, message: str, headers: Optional[Dict] = None) -> Exception:
        return NotImplementedError()  # pragma: no cover



from ._app_helper import AppHelper
from ._attribute import get_app_attribute, set_app_attribute
from ._exception import http_exception
from ._load_app import load_app
from ._pait import Pait, pait
from ._simple_route import SimpleRoute, add_multi_simple_route, add_simple_route
from ._test_helper import StarletteTestHelper, TestHelper



from abc import ABCMeta
from typing import Callable, Optional, Union

from any_api.openapi.model.openapi import security

from pait.field import Cookie, Header, Query

from .base import BaseSecurity
from .util import set_and_check_field

APIKEY_FIELD_TYPE = Union[Query, Header, Cookie]


class BaseAPIKey(BaseSecurity, metaclass=ABCMeta):
    def __init__(
        self,
        *,
        name: str,
        field: APIKEY_FIELD_TYPE,
        verify_api_key_callable: Optional[Callable[[str], bool]] = None,
        security_name: Optional[str] = None,
    ) -> None:
        field_name: str = field.get_field_name()
        if field_name not in ("query", "header", "cookie"):
            raise ValueError(f"APIKey not support {field}")
        self.not_authenticated_exc: Exception = self.get_exception(status_code=403, message="Not authenticated")
        set_and_check_field(field, name, self.not_authenticated_exc)

        self.verify_api_key_callable: Optional[Callable[[str], bool]] = verify_api_key_callable
        self.model: security.ApiKeySecurityModel = security.ApiKeySecurityModel(
            name=name, in_stub=field_name  # type: ignore
        )
        self.security_name = security_name or self.__class__.__name__

        def pait_handler(authorization: str = field) -> str:
            return self.authorization_handler(authorization)

        self.set_pait_handler(pait_handler)

    def __call__(self, authorization: str = Query.i()) -> str:
        raise RuntimeError("should not call this method")  # pragma: no cover

    def authorization_handler(self, authorization: str) -> str:
        if self.verify_api_key_callable and not self.verify_api_key_callable(authorization):
            raise self.not_authenticated_exc
        return authorization



from importlib import import_module

from pait.app.auto_load_app import auto_load_app_class
from pait.plugin.cache_response import CacheResponsePlugin as _CacheResponsePlugin

__all__ = ["CacheResponsePlugin"]

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower()
CacheResponsePlugin: "_CacheResponsePlugin" = getattr(
    import_module(pait_app_path + ".plugin.cache_response"), "CacheResponsePlugin"
)



from typing import Optional, Tuple

from pait.field import BaseField


def set_and_check_field(pait_field: BaseField, alias: str, not_authenticated_exc: Optional[Exception] = None) -> None:
    if pait_field.alias is not None:
        raise ValueError("Custom alias parameters are not allowed")
    if pait_field.not_value_exception is not None:
        raise ValueError("Custom not_value_exception parameters are not allowed")
    pait_field.set_alias(alias)
    if not_authenticated_exc:
        pait_field.not_value_exception = not_authenticated_exc


def get_authorization_scheme_param(authorization_header_value: str) -> Tuple[str, str]:
    if not authorization_header_value:
        return "", ""
    scheme, _, param = authorization_header_value.partition(" ")
    return scheme, param



from pait.app.base.security.oauth2 import (
    BaseOAuth2PasswordBearer,
    BaseOAuth2PasswordRequestFrom,
    OAuth2PasswordBearerJsonRespModel,
    OAuth2PasswordRequestFrom,
    OAuth2PasswordRequestFromStrict,
)

from .util import GetException

__all__ = [
    "OAuth2PasswordBearer",
    "OAuth2PasswordRequestFrom",
    "OAuth2PasswordRequestFromStrict",
    "BaseOAuth2PasswordRequestFrom",
    "OAuth2PasswordBearerJsonRespModel",
]


class OAuth2PasswordBearer(GetException, BaseOAuth2PasswordBearer):
    pass



from typing import Any, Type

from starlette.responses import JSONResponse, Response

from pait.model import BaseResponseModel, JsonResponseModel


def gen_response(response_value: Any, response_model_class: Type[BaseResponseModel], *args: Any, **kwargs: Any) -> Any:
    if isinstance(response_value, Response):
        return response_value
    elif issubclass(response_model_class, JsonResponseModel):
        resp: Response = JSONResponse(response_value)
    else:
        resp = Response(response_value, media_type=response_model_class.media_type)

    resp.status_code = response_model_class.status_code[0]
    if response_model_class.header:
        resp.headers.update(response_model_class.get_header_example_dict())
    return resp


def set_info_to_response(resp: Response, response_model_class: Type[BaseResponseModel]) -> None:
    resp.media_type = response_model_class.media_type
    resp.status_code = response_model_class.status_code[0]
    if response_model_class.header:
        resp.headers.update(response_model_class.get_header_example_dict())



from typing import TYPE_CHECKING, Any

from pait.model.status import PaitStatus

if TYPE_CHECKING:
    from pait.model import Config, PaitCoreModel


__all__ = ["sync_config_data_to_pait_core_model"]


def sync_config_data_to_pait_core_model(config: "Config", pait_core_model: "PaitCoreModel", **kwargs: Any) -> None:
    """Synchronize the config data to the corresponding pait core model"""
    if not pait_core_model.author:
        pait_core_model.author = config.author
    if not pait_core_model.status or pait_core_model.status is PaitStatus.undefined:
        pait_core_model.status = config.status
    if not pait_core_model.response_model_list:
        pait_core_model.add_response_model_list(config.default_response_model_list)
    if not pait_core_model.pydantic_model_config:
        pait_core_model.pydantic_model_config = config.default_pydantic_model_config

    if config.apply_func_list:
        for apply_func in config.apply_func_list:
            apply_func(pait_core_model)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pait',
 'pait.app',
 'pait.app.any',
 'pait.app.any.plugin',
 'pait.app.any.security',
 'pait.app.base',
 'pait.app.base.adapter',
 'pait.app.base.security',
 'pait.app.flask',
 'pait.app.flask.adapter',
 'pait.app.flask.plugin',
 'pait.app.flask.security',
 'pait.app.sanic',
 'pait.app.sanic.adapter',
 'pait.app.sanic.plugin',
 'pait.app.sanic.security',
 'pait.app.starlette',
 'pait.app.starlette.adapter',
 'pait.app.starlette.plugin',
 'pait.app.starlette.security',
 'pait.app.tornado',
 'pait.app.tornado.adapter',
 'pait.app.tornado.plugin',
 'pait.app.tornado.security',
 'pait.extra',
 'pait.model',
 'pait.openapi',
 'pait.param_handle',
 'pait.plugin',
 'pait.util']

package_data = \
{'': ['*']}

install_requires = \
['any-api>=0.1.0.4,<0.2.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'typing-extensions>=4.1.1,<5.0.0']

extras_require = \
{'all': ['redis>=4.2.2,<5.0.0'], 'redis': ['redis>=4.2.2,<5.0.0']}

setup_kwargs = {
    'name': 'pait',
    'version': '1.0.0a7',
    'description': 'Pait is a Python api tool. Pait enables your Python web framework to have type checking, parameter type conversion, interface document generation and can display your documents through Redoc or Swagger (power by inspect, pydantic)',
    'long_description': '![](https://cdn.jsdelivr.net/gh/so1n/so1n_blog_photo@master/blog_photo/1652600629491%E6%9C%AA%E5%91%BD%E5%90%8D.jpg)\n<p align="center">\n    Pait(π tool) - <em>Python Modern API Tools, easier to use web frameworks/write API routing</em>\n</p>\n<p align="center">\n    <a href="https://codecov.io/gh/so1n/pait" target="_blank">\n        <img src="https://codecov.io/gh/so1n/pait/branch/master/graph/badge.svg?token=NEVM1VODHR" alt="Coverage">\n    </a>\n</p>\n<p align="center">\n    <a href="https://pypi.org/project/pait/" target="_blank">\n        <img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/pait">\n    </a>\n    <a href="https://pypi.org/project/pait/" target="_blank">\n        <img alt="PyPI" src="https://img.shields.io/pypi/v/pait">\n    </a>\n</p>\n<p align="center">\n    <a href="https://github.com/so1n/pait/actions?query=event%3Apush+branch%3Amaster" target="_blank">\n        <img alt="GitHub Workflow Status" src="https://img.shields.io/github/actions/workflow/status/so1n/pait/python-package.yml">\n    </a>\n    <a href="https://github.com/so1n/pait/releases" target="_blank">\n        <img alt="GitHub release (release name instead of tag name)" src="https://img.shields.io/github/v/release/so1n/pait?include_prereleases">\n    </a>\n    <a href="https://github.com/so1n/pait/actions?query=event%3Apush+branch%3Amaster" target="_blank">\n        <img src="https://github.com/so1n/pait/actions/workflows/python-package.yml/badge.svg?event=push&branch=master" alt="Test">\n    </a>\n</p>\n<p align="center">\n    <a href="https://github.com/so1n/pait/tree/master/example" target="_blank">\n        <img src="https://img.shields.io/badge/Support%20framework-Flask%2CSanic%2CStarlette%2CTornado-brightgreen" alt="Support framework">\n    </a>\n</p>\n\n\n---\n**Documentation**: [https://so1n.me/pait/](https://so1n.me/pait/)\n\n**中文文档**: [https://so1n.me/pait-zh-doc/](https://so1n.me/pait-zh-doc/)\n\n---\n\n# Warning\nThere are changes between the current version and the 0.8 version of the API, For more information, please refer to [1.0.0version change](https://github.com/so1n/pait/blob/master/CHANGELOG.md)\n\n# pait\n\nPait is an api tool that can be used in any python web framework, the features provided are as follows:\n - [x] Parameter checksum automatic conversion (parameter check depends on `Pydantic`)\n - [x] Parameter dependency verification\n - [x] Automatically generate openapi files\n - [x] Swagger, Redoc, Rapidoc, Elements, OpenAPI route\n - [x] gRPC Gateway route\n - [x] TestClient support, support response result verification\n - [x] Support for plugin extensions, such as the Mock plugin, CheckResponse Plugin, Cache Response Plugin\n - [ ] WebSocket support\n - [ ] Auto API Test support\n\n\n> Note:\n>\n> - mypy check 100%\n>\n> - python version >= 3.7 (support postponed annotations)\n\n\n\n# Installation\n```Bash\npip install pait --pre\n```\nIf want to use the gRPC gateway feature, need to add a \'grpc\' dependency:\n```bash\npip install pait[grpc] --pre\n```\n# Simple Example\n```python\nfrom typing import Type\nimport uvicorn  # type: ignore\nfrom starlette.applications import Starlette\nfrom starlette.responses import JSONResponse\nfrom starlette.routing import Route\n\nfrom pait.app.starlette import pait\nfrom pait.field import Body\nfrom pait.openapi.doc_route import add_doc_route\nfrom pait.model.response import JsonResponseModel\nfrom pydantic import BaseModel, Field\n\n\nclass DemoResponseModel(JsonResponseModel):\n    """demo post api response model"""\n    class ResponseModel(BaseModel):\n        uid: int = Field()\n        user_name: str = Field()\n\n    description: str = "demo response"\n    response_data: Type[BaseModel] = ResponseModel\n\n\n@pait(response_model_list=[DemoResponseModel])\nasync def demo_post(\n    uid: int = Body.i(description="user id", gt=10, lt=1000),\n    user_name: str = Body.i(description="user name", min_length=2, max_length=4)\n) -> JSONResponse:\n    return JSONResponse({\'uid\': uid, \'user_name\': user_name})\n\n\napp = Starlette(routes=[Route(\'/api\', demo_post, methods=[\'POST\'])])\nadd_doc_route(app)\nuvicorn.run(app)\n```\nSee [documentation](https://so1n.me/pait/) for more features\n\n# Support Web framework\n\n| Framework | Description            |\n|-----------|------------------------|\n| Flask     | All features supported |\n| Sanic     | All features supported |\n| Starlette | All features supported |\n| Tornado   | All features supported |\n| Django    | Coming soon            |\n\n\nIf the web framework is not supported(which you are using).\n\nCan be modified sync web framework according to [pait.app.flask](https://github.com/so1n/pait/blob/master/pait/app/flask.py)\n\nCan be modified async web framework according to [pait.app.starlette](https://github.com/so1n/pait/blob/master/pait/app/starlette.py)\n\n# Performance\nFor the parameter validation and transformation feature, Pait is mainly responsible for injecting the request data dependency into the `Pydantic` model, and then the `Pydanitc` verifies and transforms the data.\n\nThe time consumed by this process <=0.0003 (s), for benchmarks data and subsequent optimization, see [#27](https://github.com/so1n/pait/issues/27)\n\n\n# Full example\nFor more complete examples, please refer to [example](https://github.com/so1n/pait/tree/master/example)\n',
    'author': 'So1n',
    'author_email': 'so1n897046026@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/so1n/pait',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import binascii
import warnings
from base64 import b64decode
from typing import Callable, Optional

from any_api.openapi.model.openapi.security import HttpSecurityModel
from pydantic import BaseModel

from pait.field import Header

from .base import BaseSecurity
from .util import get_authorization_scheme_param, set_and_check_field


class HTTPBasicCredentials(BaseModel):
    username: str
    password: str
    realm: Optional[str] = None


class BaseHTTPBasic(BaseSecurity):
    def __init__(
        self,
        *,
        security_model: HttpSecurityModel = HttpSecurityModel(scheme="basic"),
        security_name: Optional[str] = None,
        header_field: Optional[Header] = None,
        realm: Optional[str] = None,
        is_raise: bool = True,
    ) -> None:
        """
        :param security_model: http basic auth security model
        :param security_name: name of http basic auth security model in openapi spec
        :param header_field: pait header field
        :param realm: realm of http basic auth security model
        :param is_raise: is raise exception if there is invalid http basic auth data

        Note: If the validation fails and 'is raise' is true, only a 401 exception will be thrown,
            and if you want to throw a 403 exception, you need to handle it through inheritance or other methods
        """
        if security_model.scheme != "basic":
            raise ValueError(f"{self.__class__.__name__} only support basic scheme")  # pragma: no cover
        self.model = security_model
        self.is_raise: bool = is_raise
        self.security_name = security_name or self.__class__.__name__
        self.header_field: Header = header_field or Header.i(openapi_include=False)
        self.realm: Optional[str] = realm
        self.not_authorization_exc: Exception = self.get_exception(
            status_code=401,
            message="Not authentication",
            headers={"WWW-Authenticate": f'Basic realm="{realm}"'} if realm else {"WWW-Authenticate": "Basic"},
        )
        set_and_check_field(self.header_field, "Authorization", self.not_authorization_exc if is_raise else None)

        def pait_handler(authorization: str = self.header_field) -> Optional[HTTPBasicCredentials]:
            return self.authorization_handler(authorization)

        self.set_pait_handler(pait_handler)

    def __call__(self, authorization: str = Header.i()) -> Optional[HTTPBasicCredentials]:
        raise RuntimeError("should not call this method")  # pragma: no cover

    def authorization_handler(self, authorization: str) -> Optional[HTTPBasicCredentials]:
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or (scheme.lower() != "basic"):
            if self.is_raise:
                raise self.not_authorization_exc
            else:
                return None

        try:
            data = b64decode(param).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error):
            raise self.not_authorization_exc
        username, separator, password = data.partition(":")
        if not separator:
            raise self.not_authorization_exc
        return HTTPBasicCredentials(username=username, password=password, realm=self.realm)


class BaseHTTP(BaseSecurity):
    def __init__(
        self,
        *,
        security_model: HttpSecurityModel,
        security_name: Optional[str] = None,
        header_field: Optional[Header] = None,
        is_raise: bool = True,
        verify_callable: Optional[Callable[[str], bool]] = None,
    ):
        """
        :param security_model: http basic auth security model
        :param security_name: name of http basic auth security model in openapi spec
        :param header_field: pait header field
        :param is_raise: is raise exception if there is invalid http basic auth data
        :param verify_callable: callable to verify http basic auth data
        """
        self.model: HttpSecurityModel = security_model
        self.security_name = security_name or self.__class__.__name__
        self.header_field: Header = header_field or Header.i(openapi_include=False)
        self.is_raise: bool = is_raise
        self.verify_callable: Optional[Callable[[str], bool]] = verify_callable

        self.not_authenticated_exc: Exception = self.get_exception(status_code=403, message="Not authenticated")
        set_and_check_field(self.header_field, "Authorization", self.not_authenticated_exc if is_raise else None)

        def pait_handler(authorization: str = self.header_field) -> Optional[str]:
            return self.authorization_handler(authorization)

        self.set_pait_handler(pait_handler)

    def __call__(self, authorization: str = Header.i()) -> Optional[str]:
        raise RuntimeError("should not call this method")  # pragma: no cover

    def authorization_handler(self, authorization: str) -> Optional[str]:
        scheme, credentials = get_authorization_scheme_param(authorization)
        if (not (authorization and scheme and credentials)) or (scheme.lower() != self.model.scheme):
            if self.is_raise:
                raise self.not_authenticated_exc
            else:
                return None
        if self.verify_callable and not self.verify_callable(authorization):
            raise self.not_authenticated_exc
        return credentials


class BaseHTTPBearer(BaseHTTP):
    def __init__(
        self,
        *,
        bearer_format: Optional[str] = None,
        security_name: Optional[str] = None,
        header_field: Optional[Header] = None,
        is_raise: bool = True,
        verify_callable: Optional[Callable[[str], bool]] = None,
    ):
        """
        :param bearer_format: 'Bearer' style token, 'Authorization: Bearer <token>'
        :param security_name: name of http basic auth security model in openapi spec
        :param header_field: pait header field
        :param is_raise: is raise exception if there is invalid http basic auth data
        :param verify_callable: callable to verify http basic auth data
        """
        super().__init__(
            security_model=HttpSecurityModel(scheme="bearer", bearerFormat=bearer_format),
            security_name=security_name,
            header_field=header_field,
            is_raise=is_raise,
            verify_callable=verify_callable,
        )


class BaseHTTPDigest(BaseHTTP):
    def __init__(
        self,
        security_name: Optional[str] = None,
        header_field: Optional[Header] = None,
        is_raise: bool = True,
        verify_callable: Optional[Callable[[str], bool]] = None,
    ) -> None:
        """
        :param security_name: name of http basic auth security model in openapi spec
        :param header_field: pait header field
        :param is_raise: is raise exception if there is invalid http basic auth data
        :param verify_callable: callable to verify http basic auth data
        """
        super().__init__(
            security_model=HttpSecurityModel(scheme="digest"),
            security_name=security_name,
            header_field=header_field,
            is_raise=is_raise,
            verify_callable=verify_callable,
        )

    def authorization_handler(self, authorization: str) -> Optional[str]:
        warnings.warn("Http Digest just a simple example and does not provide practical usage")
        return super().authorization_handler(authorization)



from typing import Any, Type

from sanic.response import BaseHTTPResponse, HTTPResponse, json

from pait.model import BaseResponseModel, JsonResponseModel


def gen_response(response_value: Any, response_model_class: Type[BaseResponseModel], *args: Any, **kwargs: Any) -> Any:
    if isinstance(response_value, BaseHTTPResponse):
        return response_value
    elif issubclass(response_model_class, JsonResponseModel):
        resp: BaseHTTPResponse = json(response_value)
    else:
        resp = HTTPResponse(response_value)
        resp.content_type = response_model_class.media_type

    resp.status = response_model_class.status_code[0]
    if response_model_class.header:
        resp.headers.update(response_model_class.get_header_example_dict())
    return resp


def set_info_to_response(resp: BaseHTTPResponse, response_model_class: Type[BaseResponseModel]) -> None:
    resp.content_type = response_model_class.media_type
    resp.status = response_model_class.status_code[0]
    if response_model_class.header:
        resp.headers.update(response_model_class.get_header_example_dict())



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from pait.exceptions import CheckValueError
from pait.field import BaseField, ExtraParam
from pait.plugin.base import PluginContext, PluginManager, PostPluginProtocol
from pait.util import FuncSig, gen_tip_exc, get_func_sig

if TYPE_CHECKING:
    from pait.model.core import PaitCoreModel

__all__ = ["AtMostOneOfPlugin", "AtMostOneOfExtraParam"]


class AtMostOneOfExtraParam(ExtraParam):
    group: str


class AtMostOneOfPlugin(PostPluginProtocol):
    """
    Check whether each group of parameters appear at the same time
    """

    at_most_one_of_list: List[List[str]]

    def check_param(self, context: PluginContext) -> None:
        try:
            for at_most_one_of in self.at_most_one_of_list:
                if len([i for i in at_most_one_of if context.kwargs.get(i, None) is not None]) > 1:
                    raise CheckValueError(f"requires at most one of param {' or '.join(at_most_one_of)}")
        except Exception as e:
            raise e from gen_tip_exc(context.pait_core_model.func, e)

    @classmethod
    def pre_load_hook(cls, pait_core_model: "PaitCoreModel", kwargs: Dict) -> Dict:
        kwargs = super().pre_load_hook(pait_core_model, kwargs)
        fun_sig: FuncSig = get_func_sig(pait_core_model.func)
        at_most_one_of_dict: Dict[str, List[str]] = {}
        for param in fun_sig.param_list:
            default: Any = param.default
            if not isinstance(default, BaseField):
                continue
            for extra_param in default.extra_param_list:
                if not isinstance(extra_param, AtMostOneOfExtraParam):
                    continue
                at_most_one_of_dict.setdefault(extra_param.group, [])
                at_most_one_of_dict[extra_param.group].append(default.alias or param.name)
        for _, v in at_most_one_of_dict.items():
            kwargs["at_most_one_of_list"].append(v)
        return kwargs

    @classmethod
    def build(cls, *, at_most_one_of_list: Optional[List[List[str]]] = None) -> "PluginManager":  # type: ignore
        return super().build(at_most_one_of_list=at_most_one_of_list or [])

    def __call__(self, context: PluginContext) -> Any:
        self.check_param(context)
        return super().__call__(context)



# flake8: noqa: F403
from .config import *  # type: ignore
from .context import *  # type: ignore
from .core import *  # type: ignore
from .response import *  # type: ignore
from .status import *  # type: ignore
from .tag import *  # type: ignore
from .template import *  # type: ignore

__all__ = [
    config.__all__  # type: ignore
    + context.__all__  # type: ignore
    + core.__all__  # type: ignore
    + response.__all__  # type: ignore
    + status.__all__  # type: ignore
    + tag.__all__  # type: ignore
    + template.__all__  # type: ignore
]



import inspect
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Type, TypeVar

from pydantic import BaseConfig, BaseModel

from pait.app.base import BaseAppHelper
from pait.extra.util import sync_config_data_to_pait_core_model
from pait.field import BaseField
from pait.g import config, pait_context, pait_data
from pait.model import BaseResponseModel, ContextModel, PaitCoreModel, PaitStatus, Tag
from pait.param_handle import AsyncParamHandler, ParamHandler
from pait.plugin.base import PluginManager, PluginProtocol, PostPluginProtocol, PrePluginProtocol
from pait.util import get_func_sig

if TYPE_CHECKING:
    from param_handle import BaseParamHandler
_AppendT = TypeVar("_AppendT", list, tuple)
_PaitT = TypeVar("_PaitT", bound="Pait")
_PluginT = TypeVar("_PluginT", bound="PluginProtocol")


class Pait(object):
    app_helper_class: "Type[BaseAppHelper]"
    param_handler_plugin_class: Type[ParamHandler] = ParamHandler
    async_param_handler_plugin_class: Type[AsyncParamHandler] = AsyncParamHandler

    def __init__(
        self: "_PaitT",
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        pydantic_basemodel: Optional[Type[BaseModel]] = None,
        default_field_class: Optional[Type[BaseField]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[Tag, ...]] = None,
        response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        # plugin
        plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        param_handler_plugin: Optional[Type["BaseParamHandler"]] = None,
        **kwargs: Any,
    ):
        """
        :param pydantic_model_config: pydantic.BaseConfig
        :param pydantic_basemodel: pydantic.BaseModel
        :param default_field_class: pait.field.BaseField
        :param pre_depend_list:  List of depend functions to execute before route functions
        :param author:  The author who wrote this routing function
        :param desc:  Description of the routing function
        :param summary:  Introduction to Routing Functions
        :param name:  The name of the routing function, defaults to the function name
        :param status:  The state of the routing function
        :param group:  The group to which the routing function belongs
        :param tag:  A collection of labels for routing functions
        :param response_model_list: The response object of the route function
        :param plugin_list: pre plugin for routing functions
        :param post_plugin_list: post plugin list for routing functions
        :param param_handler_plugin: The param handler plugin of the routing function,
            the default is pait.param_handler.x
        """

        check_cls_param_list: List[str] = ["app_helper_class"]
        for cls_param in check_cls_param_list:
            if not getattr(self, cls_param, None):
                raise ValueError(
                    f"Please specify the value of the {cls_param} parameter, you can refer to `pait.app.xxx`"
                )
        if not isinstance(self.app_helper_class, type):
            raise TypeError(f"{self.app_helper_class} must be class")
        if not issubclass(self.app_helper_class, BaseAppHelper):
            raise TypeError(f"{self.app_helper_class} must sub from {BaseAppHelper.__class__.__name__}")

        # Note: pydantic not check field.default value when config.validate_all = False
        # See issue: https://github.com/samuelcolvin/pydantic/issues/1280
        self._pydantic_model_config: Optional[Type[BaseConfig]] = pydantic_model_config
        self._pydantic_basemodel: Optional[Type[BaseModel]] = pydantic_basemodel
        self._default_field_class: Optional[Type[BaseField]] = default_field_class
        # param check
        self._pre_depend_list: Optional[List[Callable]] = pre_depend_list
        # doc
        self._author: Optional[Tuple[str, ...]] = author
        self._desc: Optional[str] = desc
        self._summary: Optional[str] = summary
        self._name: Optional[str] = name
        self._status: Optional[PaitStatus] = status
        self._group: Optional[str] = group
        self._tag: Optional[Tuple[Tag, ...]] = tag
        self._response_model_list: Optional[List[Type[BaseResponseModel]]] = response_model_list
        # plugin
        self._plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = plugin_list
        self._post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = post_plugin_list
        self._param_handler_plugin: Optional[Type["BaseParamHandler"]] = param_handler_plugin
        self.extra: dict = kwargs

    @staticmethod
    def _append_data(
        key: str,
        target_container: Optional[_AppendT],
        append_container: Optional[_AppendT],
        self_container: Optional[_AppendT],
    ) -> Optional[_AppendT]:
        if target_container and append_container:
            raise KeyError(f"{key} and append_{key} cannot be used together")  # pragma: no cover
        if append_container:
            return (self_container or append_container.__class__()) + append_container
        elif target_container is None:
            return self_container
        else:
            return target_container

    @property
    def response_model_list(self) -> List[Type[BaseResponseModel]]:
        return self._response_model_list or []

    def create_sub_pait(
        self: "_PaitT",
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        pydantic_basemodel: Optional[Type[BaseModel]] = None,
        default_field_class: Optional[Type[BaseField]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        append_pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        append_author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[Tag, ...]] = None,
        append_tag: Optional[Tuple[Tag, ...]] = None,
        response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        append_response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        append_plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        append_post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        param_handler_plugin: Optional[Type["BaseParamHandler"]] = None,
        **kwargs: Any,
    ) -> _PaitT:
        """
        :param pydantic_model_config: pydantic.BaseConfig
        :param pydantic_basemodel: pydantic.BaseModel
        :param default_field_class: pait.field.BaseField
        :param pre_depend_list:  List of depend functions to execute before route functions(
            Do not use the pre depend value specified when Pait is initialized)
        :param append_pre_depend_list: Append some author when creating child Pait
        :param author:  The author who wrote this routing function(
            Do not use the author value specified when Pait is initialized)
        :param append_author: Append some author when creating child Pait
        :param desc:  Description of the routing function
        :param summary:  Introduction to Routing Functions
        :param name:  The name of the routing function, defaults to the function name
        :param status:  The state of the routing function
        :param group:  The group to which the routing function belongs
        :param tag:  A collection of labels for routing functions(
            Do not use the tag value specified when Pait is initialized)
        :param append_tag: Append some tags when creating child Pait
        :param response_model_list: The response object of the route function(
            Do not use the response model value specified when Pait is initialized)
        :param append_response_model_list: Append some response object when creating child Pait
        :param plugin_list: pre plugin for routing functions(
            Do not use the pre plugin value specified when Pait is initialized)
        :param append_plugin_list: Append some pre plugin when creating child Pait
        :param post_plugin_list: post plugin list for routing functions(
            Do not use the post plugin value specified when Pait is initialized)
        :param append_post_plugin_list:  Append some post plugin when creating child Pait
        :param param_handler_plugin: The param handler plugin of the routing function,
         the default is pait.param_handler.x
        """
        pre_depend_list = self._append_data(
            "pre_depend_list", pre_depend_list, append_pre_depend_list, self._pre_depend_list
        )
        author = self._append_data("author", author, append_author, self._author)
        tag = self._append_data("tag", tag, append_tag, self._tag)
        response_model_list = self._append_data(
            "response_model_list", response_model_list, append_response_model_list, self._response_model_list
        )
        plugin_list = self._append_data("plugin_list", plugin_list, append_plugin_list, self._plugin_list)
        post_plugin_list = self._append_data(
            "post_plugin_list", post_plugin_list, append_post_plugin_list, self._post_plugin_list
        )

        return self.__class__(
            pydantic_model_config=pydantic_model_config or self._pydantic_model_config,
            pydantic_basemodel=pydantic_basemodel or self._pydantic_basemodel,
            default_field_class=default_field_class or self._default_field_class,
            desc=desc or self._desc,
            summary=summary or self._summary,
            name=name or self._name,
            status=status or self._status,
            pre_depend_list=pre_depend_list,
            author=author,
            group=group,
            tag=tag,
            response_model_list=response_model_list,
            plugin_list=plugin_list,
            post_plugin_list=post_plugin_list,
            param_handler_plugin=param_handler_plugin or self._param_handler_plugin,
            **(kwargs or self.extra),
        )

    @staticmethod
    def init_context(pait_core_model: "PaitCoreModel", args: Any, kwargs: Any) -> ContextModel:
        """Inject App Helper into context"""
        app_helper: BaseAppHelper = pait_core_model.app_helper_class(args, kwargs)
        context: ContextModel = ContextModel(
            cbv_instance=app_helper.cbv_instance,
            app_helper=app_helper,
            pait_core_model=pait_core_model,
            args=args,
            kwargs=kwargs,
        )
        pait_context.set(context)
        return context

    def __call__(
        self: "_PaitT",
        pydantic_model_config: Optional[Type[BaseConfig]] = None,
        pydantic_basemodel: Optional[Type[BaseModel]] = None,
        default_field_class: Optional[Type[BaseField]] = None,
        # param check
        pre_depend_list: Optional[List[Callable]] = None,
        append_pre_depend_list: Optional[List[Callable]] = None,
        # doc
        author: Optional[Tuple[str, ...]] = None,
        append_author: Optional[Tuple[str, ...]] = None,
        desc: Optional[str] = None,
        summary: Optional[str] = None,
        name: Optional[str] = None,
        status: Optional[PaitStatus] = None,
        group: Optional[str] = None,
        tag: Optional[Tuple[Tag, ...]] = None,
        append_tag: Optional[Tuple[Tag, ...]] = None,
        response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        append_response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
        # plugin
        plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        append_plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
        post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        append_post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
        param_handler_plugin: Optional[Type["BaseParamHandler"]] = None,
        feature_code: str = "",
        **kwargs: Any,
    ) -> Callable:
        """
        :param pydantic_model_config: pydantic.BaseConfig
        :param pydantic_basemodel: pydantic.BaseModel
        :param default_field_class: pait.field.BaseField
        :param pre_depend_list:  List of depend functions to execute before route functions
        :param author:  The author who wrote this routing function
        :param desc:  Description of the routing function
        :param summary:  Introduction to Routing Functions
        :param name:  The name of the routing function, defaults to the function name
        :param status:  The state of the routing function
        :param group:  The group to which the routing function belongs
        :param tag:  A collection of labels for routing functions(
            Do not use the tag value specified when Pait is initialized)
        :param append_tag: Append some tags to the routing function
        :param response_model_list: The response object of the route function(
            Do not use the response_model_list value specified when Pait is initialized)
        :param append_response_model_list: Append some response object to the routing function
        :param plugin_list: pre plugin for routing functions
        :param append_plugin_list: Append some pre plugin to the routing function (
            Do not use the pre plugin value specified when Pait is initialized)
        :param post_plugin_list: post plugin list for routing functions
        :param append_post_plugin_list:  Append some post plugin  to the routing function (
            Do not use the post plugin value specified when Pait is initialized)
        :param param_handler_plugin: The param handler plugin of the routing function,
            the default is pait.param_handler.x
        :param feature_code: Specify the prefix of the pait id corresponding to the generated routing function.
            Usually, the pait_id is equal to md5(func), but during dynamic generation,
            there may be multiple different routing functions generated by the same func.
            In this case, different feature_code is needed to generate different pait_id(feature_code + md5(func))
        """
        app_name: str = self.app_helper_class.app_name

        pre_depend_list = self._append_data(
            "pre_depend_list", pre_depend_list, append_pre_depend_list, self._pre_depend_list
        )
        author = self._append_data("author", author, append_author, self._author)
        tag = self._append_data("tag", tag, append_tag, self._tag)
        response_model_list = self._append_data(
            "response_model_list", response_model_list, append_response_model_list, self._response_model_list
        )
        plugin_list = self._append_data("plugin_list", plugin_list, append_plugin_list, self._plugin_list)
        post_plugin_list = self._append_data(
            "post_plugin_list", post_plugin_list, append_post_plugin_list, self._post_plugin_list
        )

        def wrapper(func: Callable) -> Callable:
            # Pre-parsing function signatures
            get_func_sig(func)

            # load param handler plugin
            _param_handler_plugin = param_handler_plugin or self._param_handler_plugin
            if _param_handler_plugin is None:
                if inspect.iscoroutinefunction(func):
                    _param_handler_plugin = self.async_param_handler_plugin_class
                else:
                    _param_handler_plugin = self.param_handler_plugin_class
            # gen pait core model and register to pait data
            pait_core_model: PaitCoreModel = PaitCoreModel(
                func,
                self.app_helper_class,
                author=author,
                desc=desc or self._desc,
                summary=summary or self._summary,
                func_name=name or self._name,
                status=status or self._status,
                group=group or self._group,
                tag=tag,
                response_model_list=response_model_list,
                pre_depend_list=pre_depend_list,
                pydantic_model_config=pydantic_model_config or self._pydantic_model_config,
                pydantic_basemodel=pydantic_basemodel or self._pydantic_basemodel,
                plugin_list=plugin_list,
                post_plugin_list=post_plugin_list,
                param_handler_plugin=_param_handler_plugin,
                feature_code=feature_code,
                default_field_class=default_field_class or self._default_field_class,
                **(kwargs or self.extra),
            )
            sync_config_data_to_pait_core_model(config, pait_core_model)
            pait_data.register(app_name, pait_core_model)
            if inspect.iscoroutinefunction(func):

                @wraps(func)
                async def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    context: ContextModel = self.init_context(pait_core_model, args, kwargs)
                    return await pait_core_model.main_plugin(context)

                return dispatch
            else:

                @wraps(func)
                def dispatch(*args: Any, **kwargs: Any) -> Callable:
                    context = self.init_context(pait_core_model, args, kwargs)
                    return pait_core_model.main_plugin(context)

                return dispatch

        return wrapper



from importlib import import_module

from pait.app.auto_load_app import auto_load_app_class
from pait.plugin.mock_response import MockPluginProtocol as _MockPluginProtocol

__all__ = ["MockPlugin"]

pait_app_path: str = "pait.app." + auto_load_app_class().__name__.lower()
MockPlugin: "_MockPluginProtocol" = getattr(import_module(pait_app_path + ".plugin"), "MockPlugin")



from dataclasses import MISSING
from importlib import import_module
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type

from pydantic import BaseConfig, BaseModel
from typing_extensions import NoReturn

from pait.app.any.util import base_call_func, sniffing
from pait.app.auto_load_app import auto_load_app_class
from pait.app.base.simple_route import SimpleRoute
from pait.field import BaseField
from pait.model import BaseResponseModel, PaitCoreModel, PaitStatus, Tag
from pait.plugin.base import PluginManager, PostPluginProtocol, PrePluginProtocol

if TYPE_CHECKING:
    from pait.openapi.openapi import OpenAPI
    from pait.param_handle import BaseParamHandler


_NotFoundFrameworkException = RuntimeError(
    "The web framework to use cannot be found automatically, please specify it manually"
)


class Empty(object):
    def __int__(self) -> NoReturn:
        raise _NotFoundFrameworkException  # pragma: no cover

    def __setattr__(self, key: Any, value: Any) -> NoReturn:
        raise _NotFoundFrameworkException  # pragma: no cover

    def __getattr__(self, item: Any) -> NoReturn:
        raise _NotFoundFrameworkException  # pragma: no cover

    def __call__(self, *args: Any, **kwargs: Any) -> NoReturn:
        raise _NotFoundFrameworkException  # pragma: no cover


try:
    load_class_app: Any = auto_load_app_class()
    pait_app_path: str = "pait.app." + load_class_app.__name__.lower()
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from pait.core import Pait as _Pait

    Pait: "_Pait" = getattr(import_module(pait_app_path), "Pait")
    http_exception = getattr(import_module(pait_app_path), "http_exception")
except RuntimeError:  # pragma: no cover
    # Automatic loading of classes, loading failure when the user can not use
    load_class_app = Empty()
    Pait = Empty()  # type: ignore
    pait_app_path = ""  # pragma: no cover
    http_exception = Empty()  # type: ignore


def load_app(app: Any, auto_load_route: bool = False) -> Dict[str, PaitCoreModel]:
    """Read data from the route that has been registered to `pait`
    Note:This is an implicit method
    """
    return base_call_func("load_app", app, app=app, auto_load_route=auto_load_route)


def pait(
    pydantic_model_config: Optional[Type[BaseConfig]] = None,
    pydantic_basemodel: Optional[Type[BaseModel]] = None,
    default_field_class: Optional[Type[BaseField]] = None,
    # param check
    pre_depend_list: Optional[List[Callable]] = None,
    append_pre_depend_list: Optional[List[Callable]] = None,
    # doc
    author: Optional[Tuple[str, ...]] = None,
    append_author: Optional[Tuple[str, ...]] = None,
    desc: Optional[str] = None,
    summary: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[PaitStatus] = None,
    group: Optional[str] = None,
    tag: Optional[Tuple[Tag, ...]] = None,
    append_tag: Optional[Tuple[Tag, ...]] = None,
    response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
    append_response_model_list: Optional[List[Type[BaseResponseModel]]] = None,
    # plugin
    plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
    append_plugin_list: Optional[List[PluginManager[PrePluginProtocol]]] = None,
    post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
    append_post_plugin_list: Optional[List[PluginManager[PostPluginProtocol]]] = None,
    param_handler_plugin: Optional[Type["BaseParamHandler"]] = None,
    feature_code: str = "",
    **kwargs: Any,
) -> Callable:
    """provide parameter checks and type conversions for each routing function/cbv class
    Note:This is an implicit method
    """
    if not pait_app_path:
        raise RuntimeError("Auto load app fail")  # pragma: no cover
    _pait: Optional[Callable] = getattr(import_module(pait_app_path), "pait")
    if not _pait:
        raise NotImplementedError(f"Pait not support:{load_class_app}")
    return _pait(
        pydantic_basemodel=pydantic_basemodel,
        pydantic_model_config=pydantic_model_config,
        default_field_class=default_field_class,
        pre_depend_list=pre_depend_list,
        append_pre_depend_list=append_pre_depend_list,
        author=author,
        append_author=append_author,
        desc=desc,
        summary=summary,
        name=name,
        status=status,
        group=group,
        tag=tag,
        append_tag=append_tag,
        response_model_list=response_model_list,
        append_response_model_list=append_response_model_list,
        plugin_list=plugin_list,
        append_plugin_list=append_plugin_list,
        post_plugin_list=post_plugin_list,
        append_post_plugin_list=append_post_plugin_list,
        param_handler_plugin=param_handler_plugin,
        feature_code=feature_code,
        **kwargs,
    )


def set_app_attribute(app: Any, key: str, value: Any) -> None:
    base_call_func("set_app_attribute", app, key, value, app=app)


def get_app_attribute(app: Any, key: str, default_value: Any = MISSING) -> Any:
    return base_call_func("get_app_attribute", app, key, default_value, app=app)


def add_simple_route(app: Any, simple_route: "SimpleRoute") -> None:
    base_call_func("add_simple_route", app, simple_route, app=app)


def add_multi_simple_route(
    app: Any, *simple_route_list: "SimpleRoute", prefix: str = "/", title: str = "", **kwargs: Any
) -> None:
    base_call_func("add_multi_simple_route", app, *simple_route_list, app=app, prefix=prefix, title=title, **kwargs)



from dataclasses import MISSING, dataclass, field
from inspect import Parameter
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence

from pait.util import get_parameter_list_from_class

if TYPE_CHECKING:
    from pait.app.base import BaseAppHelper
    from pait.model.core import PaitCoreModel


__all__ = ["ContextModel"]


@dataclass
class ContextModel(object):
    cbv_instance: Optional[Any]
    app_helper: "BaseAppHelper"
    pait_core_model: "PaitCoreModel"
    # If it is a pre plugin,
    #   args and kwargs are the parameters for the corresponding web framework to call the route.
    # If it is a post plugin,
    #   args and kwargs are the parameters filled in by the developer to write the routing function.
    args: Sequence
    kwargs: dict

    cbv_param_list: List[Parameter] = field(init=False)
    # If it is not used, then it is not initialized, saving memory footprint
    state: Dict[str, Any] = field(init=False)

    def __post_init__(self) -> None:
        if self.cbv_instance:
            self.cbv_param_list = get_parameter_list_from_class(self.cbv_instance.__class__)
        else:
            self.cbv_param_list = []

    def _init_state(self) -> None:
        if not hasattr(self, "state"):
            self.state = {}

    def set_to_state(self, key: str, value: Any) -> None:
        self._init_state()
        self.state[key] = value

    def get_form_state(self, key: str, default_value: Any = MISSING) -> Any:
        value: Any = self.state.get(key, default_value)
        if value is MISSING:
            raise KeyError(key)
        return value



import re
from typing import Callable

from flask import Blueprint, Flask

from pait.app.base.simple_route import SimpleRoute, add_route_plugin
from pait.app.flask.plugin.unified_response import UnifiedResponsePlugin

__all__ = ["SimpleRoute", "add_simple_route", "add_multi_simple_route"]


def replace_openapi_url_to_url(url: str) -> str:
    """Convert the OpenAPI URL format to a format supported by the web framework

    >>> assert "http://google.com/user/{user_id}/post" == replace_openapi_url_to_url(
    >>>    "http://google.com/user/<path:user_id>/post"
    >>> )
    """
    matches = re.findall(r"{([a-zA-Z_]+)}", url)
    for match in matches:
        url = url.replace(f"{{{match}}}", f"<path:{match}>")
    return url


def add_simple_route(
    app: Flask,
    simple_route: "SimpleRoute",
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    add_route_plugin(simple_route, UnifiedResponsePlugin)
    app.add_url_rule(
        _replace_openapi_url_to_url(simple_route.url), view_func=simple_route.route, methods=simple_route.methods
    )


def add_multi_simple_route(
    app: Flask,
    *simple_route_list: "SimpleRoute",
    prefix: str = "/",
    title: str = "",
    import_name: str = "",
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    blueprint: Blueprint = Blueprint(
        title,
        import_name=import_name,
        url_prefix=prefix,
    )
    for simple_route in simple_route_list:
        add_route_plugin(simple_route, UnifiedResponsePlugin)
        blueprint.add_url_rule(
            _replace_openapi_url_to_url(simple_route.url), view_func=simple_route.route, methods=simple_route.methods
        )
    app.register_blueprint(blueprint)



import re
from typing import Callable, Type

from tornado.routing import _RuleList
from tornado.web import AnyMatches, Application, RequestHandler, Rule, _ApplicationRouter

from pait.app.base.simple_route import SimpleRoute, add_route_plugin
from pait.app.tornado.plugin.unified_response import UnifiedResponsePlugin

__all__ = ["SimpleRoute", "add_simple_route", "add_multi_simple_route"]


def replace_openapi_url_to_url(url: str) -> str:
    """Convert the OpenAPI URL format to a format supported by the web framework"""
    matches = re.findall(r"{([a-zA-Z_]+)}", url)
    for match in matches:
        url = url.replace(f"{{{match}}}", "(?P<" + match + r">\w+)")
    return r"{}".format(url)


def _add_route(
    app: Application,
    request_handler: Type[RequestHandler],
    _replace_openapi_url_to_url: Callable[[str], str],
    *simple_route_list: "SimpleRoute",
    prefix: str = "",
    title: str = "",
) -> None:
    rule_list: _RuleList = []
    for simple_route in simple_route_list:
        add_route_plugin(simple_route, UnifiedResponsePlugin)

        if title:
            model_str: str = f"{__name__}.{title}.{simple_route.route.__name__}"
        else:
            model_str = f"{__name__}.{simple_route.route.__name__}"
        route_class = type(
            simple_route.route.__name__.title() + "Handler", (request_handler,), {"__model__": model_str}
        )

        for method in simple_route.methods:
            setattr(route_class, method.lower(), simple_route.route)

        url: str = simple_route.url
        if prefix:
            if prefix.endswith("/"):
                prefix = prefix[:-1]
            if simple_route.url.startswith("/"):
                url = simple_route.url[1:]
            url = f"{prefix}/{url}"

        rule_list.append((_replace_openapi_url_to_url(url), route_class))

    # Method 1
    # app.add_handlers(r".*$", rule_list)
    #
    # Method 2
    # app.default_router.add_rules(rule_list)
    #
    # Method 3
    app.wildcard_router.add_rules(rule_list)
    app.default_router = _ApplicationRouter(app, [Rule(AnyMatches(), app.wildcard_router)])


def add_simple_route(
    app: Application,
    simple_route: "SimpleRoute",
    request_handler: Type[RequestHandler] = RequestHandler,
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    _add_route(app, request_handler, _replace_openapi_url_to_url, simple_route)


def add_multi_simple_route(
    app: Application,
    *simple_route_list: "SimpleRoute",
    prefix: str = "/",
    title: str = "",
    request_handler: Type[RequestHandler] = RequestHandler,
    _replace_openapi_url_to_url: Callable[[str], str] = replace_openapi_url_to_url,
) -> None:
    _add_route(app, request_handler, _replace_openapi_url_to_url, *simple_route_list, prefix=prefix, title=title)



from pait.app.tornado.plugin.auto_complete_json_resp import AutoCompleteJsonRespPlugin
from pait.app.tornado.plugin.check_json_resp import CheckJsonRespPlugin
from pait.app.tornado.plugin.mock_response import MockPlugin
from pait.app.tornado.plugin.unified_response import UnifiedResponsePlugin, UnifiedResponsePluginProtocol
from pait.plugin.at_most_one_of import AtMostOneOfExtraParam, AtMostOneOfPlugin
from pait.plugin.required import RequiredExtraParam, RequiredGroupExtraParam, RequiredPlugin

__all__ = [
    "AutoCompleteJsonRespPlugin",
    "RequiredExtraParam",
    "RequiredGroupExtraParam",
    "RequiredPlugin",
    "AtMostOneOfPlugin",
    "AtMostOneOfExtraParam",
    "CheckJsonRespPlugin",
    "MockPlugin",
    "UnifiedResponsePlugin",
    "UnifiedResponsePluginProtocol",
]
